"""Utilities for connecting to modbus servers"""

from __future__ import annotations

import asyncio
import socket
from contextlib import contextmanager
from datetime import datetime, timedelta
from typing import TYPE_CHECKING

import pandas as pd
from pyModbusTCP import constants as mb_const
from pyModbusTCP.client import ModbusClient

from eta_utility import get_logger
from eta_utility.connectors.node import NodeModbus
from eta_utility.connectors.util import (
    IntervalChecker,
    RetryWaiter,
    bitarray_to_registers,
    decode_modbus_value,
    encode_bits,
)

if TYPE_CHECKING:
    from collections.abc import Generator, Mapping
    from typing import Any

    from eta_utility.type_hints import Nodes, TimeStep

from .base_classes import Connection, SubscriptionHandler

log = get_logger("connectors.modbus")


class ModbusConnection(Connection[NodeModbus], protocol="modbus"):
    """The Modbus Connection class allows reading and writing from and to Modbus servers and clients. Additionally,
    it implements a subscription service, which reads continuously in a specified interval.

    :param url: URL of the Modbus Server.
    :param usr: No login supported, only here to satisfy the interface
    :param pwd: No login supported, only here to satisfy the interface
    :param nodes: List of nodes to use for all operations.
    """

    def __init__(
        self, url: str, usr: str | None = None, pwd: str | None = None, *, nodes: Nodes[NodeModbus] | None = None
    ) -> None:
        super().__init__(url, usr, pwd, nodes=nodes)

        if self._url.scheme != "modbus.tcp":
            raise ValueError("Given URL is not a valid Modbus url (scheme: modbus.tcp)")

        self.connection: ModbusClient = ModbusClient(host=self._url.hostname, port=self._url.port, timeout=2)

        self._subscription_open: bool = False
        self._subscription_nodes: set[NodeModbus] = set()
        self._sub: asyncio.Task

        self._retry = RetryWaiter()
        self._retry_interval_checker = RetryWaiter()

        self._connection_interval_checker = IntervalChecker()

    @classmethod
    def _from_node(
        cls, node: NodeModbus, usr: str | None = None, pwd: str | None = None, **kwargs: Any
    ) -> ModbusConnection:
        """Initialize the connection object from a modubs protocol node object.

        :param node: Node to initialize from.
        :param usr: Username to use.
        :param pwd: Password to use.
        :param kwargs: Other arguments are ignored.
        :return: ModbusConnection object.
        """

        return super()._from_node(node, usr=usr, pwd=pwd)

    def read(self, nodes: Nodes[NodeModbus] | None = None) -> pd.DataFrame:
        """Read some manually selected nodes from Modbus server.

        :param nodes: List of nodes to read from.
        :return: Dictionary containing current values of the Modbus variables.
        """
        _nodes = self._validate_nodes(nodes)

        values = {}

        with self._connection():
            results = {node: self._read_mb_value(node) for node in _nodes}

        for node, result in results.items():
            value = decode_modbus_value(result, node.mb_byteorder, node.dtype, node.mb_wordorder)
            values[node.name] = value

        return pd.DataFrame(values, index=[self._assert_tz_awareness(datetime.now())])

    def write(self, values: Mapping[NodeModbus, Any]) -> None:
        """Write some manually selected values on Modbus capable controller.

        .. warning::
            This is not implemented.

        :param values: Dictionary of nodes and data to write {node: value}.
        """
        nodes = self._validate_nodes(set(values.keys()))

        with self._connection():
            for node in nodes:
                if not isinstance(values[node], list):
                    bits = encode_bits(values[node], node.mb_byteorder, node.mb_bit_length, node.dtype)
                else:
                    bits = values[node]

                self._write_mb_value(node, bits)

    def subscribe(
        self, handler: SubscriptionHandler, nodes: Nodes[NodeModbus] | None = None, interval: TimeStep = 1
    ) -> None:
        """Subscribe to nodes and call handler when new data is available. Basic architecture of the subscription is
        the client- server communication. This function works asynchronously.

        :param nodes: Identifiers for the nodes to subscribe to.
        :param handler: SubscriptionHandler object with a push method that accepts node, value pairs.
        :param interval: Interval for receiving new data. It is interpreted as seconds when given as an integer.
        """
        _nodes = self._validate_nodes(nodes)

        interval = interval if isinstance(interval, timedelta) else timedelta(seconds=interval)

        self._subscription_nodes.update(_nodes)

        if self._subscription_open:
            # Adding nodes to subscription is enough to include them in the query. Do not start an additional loop
            # if one already exists
            return

        self._subscription_open = True

        loop = asyncio.get_event_loop()
        self._sub = loop.create_task(self._subscription_loop(handler, float(interval.total_seconds())))

    def close_sub(self) -> None:
        """Close the subsription."""
        self._subscription_open = False
        if self.exc:
            raise self.exc

        try:
            self._sub.cancel()
        except Exception:
            pass

        try:
            self.connection.close()
        except Exception:
            pass

    async def _subscription_loop(self, handler: SubscriptionHandler, interval: float) -> None:
        """The subscription loop handles requesting data from the server in the specified interval.

        :param handler: Handler object with a push function to receive data.
        :param interval: Interval for requesting data in seconds.
        """

        try:
            while self._subscription_open:
                try:
                    await self._retry.wait_async()
                    self._connect()
                except ConnectionError as e:
                    log.warning(str(e))
                    continue

                for node in self._subscription_nodes:
                    result = None
                    try:
                        result = self._read_mb_value(node)
                    except ValueError as e:
                        log.warning(str(e))
                    except ConnectionError:
                        handler.push(node, pd.NA, self._assert_tz_awareness(datetime.now()))
                        continue

                    if result is not None:
                        _result = decode_modbus_value(result, node.mb_byteorder, node.dtype, node.mb_wordorder)

                        time = self._assert_tz_awareness(datetime.now())

                        handler.push(node, _result, time)

                        self._connection_interval_checker.push(node=node, value=_result, timestamp=time)

                if self.connection.is_open:
                    _changed_within_interval = self._connection_interval_checker.check_interval_connection()

                    if not _changed_within_interval:
                        log.warning(
                            f"The subscription connection for {self.url} doesn't change the values "
                            "anymore. Trying to reconnect."
                        )
                        self._retry_interval_checker.tried()
                        await self._retry_interval_checker.wait_async()
                    else:
                        self._retry_interval_checker.success()
                        await asyncio.sleep(interval)
        except BaseException as e:
            self.exc = e

    def _read_mb_value(self, node: NodeModbus) -> list[int]:
        """Read raw value from modbus server. This function should not be used directly. It does not
        establish a connection or handle connection errors.
        """
        if not self.connection.is_open:
            raise ConnectionError(f"Could not establish connection to host {self.url}")

        self.connection.unit_id = node.mb_slave

        if node.mb_register == "holding":
            result = self.connection.read_holding_registers(node.mb_channel, node.mb_bit_length // 16)
        elif node.mb_register == "coils":
            result = self.connection.read_coils(node.mb_channel, node.mb_bit_length)
        elif node.mb_register == "discrete_input":
            result = self.connection.read_discrete_inputs(node.mb_channel, node.mb_bit_length)
        elif node.mb_register == "input":
            result = self.connection.read_input_registers(node.mb_channel, node.mb_bit_length // 16)
        else:
            raise ValueError(f"The specified register type for '{node.name}' is not supported: {node.mb_register}")

        if result is None:
            self._handle_mb_error()
        else:
            result = [int(x) for x in result]
        return result

    def _write_mb_value(self, node: NodeModbus, value: list[int]) -> None:
        """Write raw value to the modbus server. This function should not be used directly. It does not establish
        a connection or handle connection errors.
        """
        if not self.connection.is_open:
            raise ConnectionError(f"Could not establish connection to host {self.url}.")

        self.connection.unit_id = node.mb_slave

        if node.mb_register == "coils":
            success = self.connection.write_multiple_coils(node.mb_channel, value)
        elif node.mb_register == "holding":
            success = self.connection.write_multiple_registers(node.mb_channel, bitarray_to_registers(value))
        else:
            raise ValueError(f"The specified register type is not supported for writing: {node.mb_register}")

        if not success:
            raise ConnectionError(f"Could not write value to channel {node.mb_channel} on server: {self.url}.")

    def _connect(self) -> None:
        """Connect to server."""
        try:
            if not self.connection.is_open:
                self._retry.tried()
                if not self.connection.open():
                    raise ConnectionError(f"Could not establish connection to host {self.url}")
        except (socket.herror, socket.gaierror) as e:
            raise ConnectionError(f"Host not found: {self.url}") from e
        except socket.timeout as e:
            raise ConnectionError(f"Host timeout: {self.url}") from e
        except (RuntimeError, ConnectionError) as e:
            raise ConnectionError(f"Connection Error: {self.url}, Error: {e!s}") from e
        else:
            if self.connection.is_open:
                self._retry.success()
            else:
                raise ConnectionError(f"Could not establish connection to host {self.url}")

    def _disconnect(self) -> None:
        """Disconnect from server."""
        try:
            self.connection.close()
        except (OSError, RuntimeError) as e:
            log.error(f"Closing connection to server {self.url} failed")
            log.info(f"Connection to {self.url} returned error: {e}")
        except AttributeError:
            log.error(f"Connection to server {self.url} already closed.")

    @contextmanager
    def _connection(self) -> Generator:
        """Connect to the server and return a context manager that automatically disconnects when finished."""
        try:
            self._connect()
            yield None
        except ConnectionError as e:
            raise e
        finally:
            self._disconnect()

    def _handle_mb_error(self) -> None:
        error = self.connection.last_error
        exception = self.connection.last_except

        if error != mb_const.MB_NO_ERR:
            raise ConnectionError(f"ModbusError {error} at {self.url}: {self.connection.last_error_as_txt}")
        if exception != mb_const.EXP_NONE:
            raise ConnectionError(f"ModbusError {exception} at {self.url}: {self.connection.last_except_as_txt}")

        raise ConnectionError(f"Unknown ModbusError at {self.url}")

    def _validate_nodes(self, nodes: Nodes[NodeModbus] | None) -> set[NodeModbus]:
        vnodes = super()._validate_nodes(nodes)
        _nodes = set()
        for node in vnodes:
            if isinstance(node, NodeModbus):
                _nodes.add(node)

        return _nodes



from .callbacks import (
    CallbackEnvironment as CallbackEnvironment,
    merge_callbacks as merge_callbacks,
)
from .common import (
    deserialize_net_arch as deserialize_net_arch,
    episode_name_string as episode_name_string,
    episode_results_path as episode_results_path,
    initialize_model as initialize_model,
    is_env_closed as is_env_closed,
    is_vectorized_env as is_vectorized_env,
    load_model as load_model,
    log_net_arch as log_net_arch,
    log_run_info as log_run_info,
    log_to_file as log_to_file,
    vectorize_environment as vectorize_environment,
)
from .extractors import CustomExtractor as CustomExtractor
from .policies import NoPolicy as NoPolicy
from .processors import (
    Fold1d as Fold1d,
    Split1d as Split1d,
)
from .schedules import LinearSchedule as LinearSchedule



from __future__ import annotations

import socket
from collections.abc import Sized
from datetime import datetime
from typing import TYPE_CHECKING

# Async import
import asyncua.sync
import pandas as pd
from asyncua import ua  # , Server as asyncServer

# Sync import
from asyncua.sync import Server, ThreadLoopNotRunning
from asyncua.ua import uaerrors

from eta_utility import ensure_timezone, get_logger, url_parse
from eta_utility.connectors.node import NodeOpcUa

if TYPE_CHECKING:
    import types
    from collections.abc import Mapping
    from typing import Any

    # Sync import
    from asyncua.sync import SyncNode as SyncOpcNode
    from typing_extensions import Self

    # Async import
    # TODO: add async import: from asyncua import Node as asyncSyncOpcNode
    # https://git.ptw.maschinenbau.tu-darmstadt.de/eta-fabrik/public/eta-utility/-/issues/270
    from eta_utility.type_hints import Nodes

log = get_logger("servers.opcua")


class OpcUaServer:
    """Provides an OPC UA server with a number of specified nodes. Each node can contain single values or arrays.

    :param namespace: Namespace of the OPC UA Server.
    :param ip: IP Address to listen on (default: None).
    :param port: Port to listen on (default: 4840).
    """

    def __init__(self, namespace: str | int, ip: str | None = None, port: int = 4840) -> None:
        #: URL of the OPC UA Server.
        self.url: str
        if ip is None:
            try:
                host = socket.gethostbyname(socket.gethostname())
            except socket.gaierror:
                host = "127.0.0.1"
            self.url = f"opc.tcp://{host}:{port}"
        else:
            self.url = f"opc.tcp://{ip}:{port}"
        log.info(f"Server Address is {self.url}")

        self._url, _, _ = url_parse(self.url)

        self._server: Server = Server()
        self._server.set_endpoint(self.url)

        self.idx: int = self._server.register_namespace(str(namespace))  #: idx: Namespace of the OPC UA _server
        log.debug(f'Server Namespace set to "{namespace}"')

        self._server.set_security_policy([ua.SecurityPolicyType.NoSecurity])
        self._server.set_server_name("ETA Utility OPC UA Server")
        self._server.start()

    def write(self, values: Mapping[NodeOpcUa, Any]) -> None:
        """Write some values directly to the OPC UA server.

        :param values: Dictionary of data to write {node.name: value}.
        """

        nodes = self._validate_nodes(set(values.keys()))

        for node in nodes:
            var = self._server.get_node(node.opc_id)
            try:
                opc_type = var.get_data_type_as_variant_type()
            except asyncua.sync.ThreadLoopNotRunning as e:
                raise ConnectionError(f"Server {self} is not running.") from e
            var.set_value(ua.Variant(values[node], opc_type))

    def read(self, nodes: Nodes[NodeOpcUa] | None = None) -> pd.DataFrame:
        """
        Read some manually selected values directly from the OPC UA server.

        :param nodes: List of nodes to read from.
        :return: pandas.DataFrame containing current values of the OPC UA-variables.
        :raises RuntimeError: When an error occurs during reading.
        """
        _nodes = self._validate_nodes(nodes)

        _dikt = {}
        for node in _nodes:
            try:
                opcua_variable = self._server.get_node(node.opc_id)
                value = opcua_variable.get_value()
                _dikt[node.name] = [value]
            except uaerrors.BadNodeIdUnknown as e:
                raise RuntimeError(
                    f"The node id ({node.opc_id}) refers to a node that does not exist in the server address space "
                    f"{self.url}. (BadNodeIdUnknown)"
                ) from e

        return pd.DataFrame(_dikt, index=[ensure_timezone(datetime.now())])

    def create_nodes(self, nodes: Nodes[NodeOpcUa]) -> None:
        """Create nodes on the server from a list of nodes. This will try to create the entire node path.

        :param nodes: List or set of nodes to create.
        """

        def create_object(parent: SyncOpcNode, child: NodeOpcUa) -> SyncOpcNode:
            children: list[SyncOpcNode] = asyncua.sync._to_sync(parent.tloop, parent.get_children())
            for obj in children:
                ident = obj.nodeid.Identifier
                ident = ident.strip() if isinstance(ident, str) else ident
                if child.opc_path_str == ident:
                    return obj
            return asyncua.sync._to_sync(parent.tloop, parent.add_object(child.opc_id, child.opc_name))

        _nodes = self._validate_nodes(nodes)

        for node in _nodes:
            try:
                if len(node.opc_path) == 0:
                    last_obj = asyncua.sync._to_sync(self._server.tloop, self._server.aio_obj.get_objects_node())
                else:
                    # Create SyncNode from asyncNode
                    sync_node = asyncua.sync._to_sync(self._server.tloop, self._server.aio_obj.get_objects_node())
                    last_obj = create_object(sync_node, node.opc_path[0])

                for key in range(1, len(node.opc_path)):
                    last_obj = create_object(last_obj, node.opc_path[key])

                init_val: Any
                if not hasattr(node, "dtype"):
                    init_val = 0.0
                elif node.dtype is int:
                    init_val = 0
                elif node.dtype is bool:
                    init_val = False
                elif node.dtype is str:
                    init_val = ""
                else:
                    init_val = 0.0

                last_obj.add_variable(node.opc_id, node.opc_name, init_val)
                log.debug(f"OPC UA Node created: {node.opc_id}")
            except uaerrors.BadNodeIdExists:
                log.warning(f"Node with NodeId : {node.opc_id} could not be created. It already exists.")
            except RuntimeError as e:
                raise ConnectionError(str(e)) from e

    def delete_nodes(self, nodes: Nodes[NodeOpcUa]) -> None:
        """Delete the given nodes and their parents (if the parents do not have other children).

        :param nodes: List or set of nodes to be deleted.
        """

        def delete_node_parents(node: SyncOpcNode, depth: int = 20) -> None:
            parents = node.get_references(direction=ua.BrowseDirection.Inverse)
            if not node.get_children():
                node.delete(delete_references=True)
                log.info(f"Deleted Node {node.nodeid} from server {self.url}.")
            else:
                log.info(f"Node {node.nodeid} on server {self.url} has remaining children and was not deleted.")
            for parent in parents:
                if depth > 0:
                    delete_node_parents(self._server.get_node(parent.NodeId), depth=depth - 1)

        nodes = self._validate_nodes(nodes)

        for node in nodes:
            delete_node_parents(self._server.get_node(node.opc_id))

    def start(self) -> None:
        """Restart the server after it was stopped."""
        self._server.start()

    def stop(self) -> None:
        """This should always be called, when the server is not needed anymore. It stops the server."""
        try:
            self._server.stop()
        except AttributeError:
            # Occurs only if server did not exist and can be ignored.
            pass
        except ThreadLoopNotRunning:
            # Occurs only if server was already stopped (and therefore the ThreadLoop as well) and can be ignored.
            pass

    @property
    def active(self) -> bool:
        return self._server.aio_obj.bserver._server._serving

    def allow_remote_admin(self, allow: bool) -> None:
        """Allow remote administration of the server.

        :param allow: Set to true to enable remote administration of the server.
        """
        self._server.aio_obj.allow_remote_admin(allow)

    def _validate_nodes(self, nodes: Nodes[NodeOpcUa] | None) -> set[NodeOpcUa]:
        """Make sure that nodes are a Set of nodes and that all nodes correspond to the protocol and url
        of the connection.

        :param nodes: Sequence of Node objects to validate.
        :return: Set of valid Node objects for this connection.
        """
        _nodes = None

        if nodes:
            if not isinstance(nodes, Sized):
                nodes = {nodes}

            # If not using preselected nodes from self.selected_nodes, check if nodes correspond to the connection
            _nodes = {
                node for node in nodes if isinstance(node, NodeOpcUa) and node.url_parsed.hostname == self._url.hostname
            }

        # Make sure that some nodes remain after the checks and raise an error if there are none.
        if not _nodes or len(_nodes) == 0:
            raise ValueError(
                f"Some nodes to read from/write to must be specified. If nodes were specified, they do not "
                f"match the connection {self.url}"
            )

        return _nodes

    def __enter__(self) -> Self:
        return self

    def __exit__(
        self, exc_type: type[BaseException] | None, exc_val: BaseException | None, exc_tb: types.TracebackType | None
    ) -> None:
        self.stop()



from __future__ import annotations

import struct
from asyncio import sleep as async_sleep
from datetime import datetime
from itertools import groupby
from time import sleep
from typing import TYPE_CHECKING

import pandas as pd

from eta_utility.connectors.node import Node
from eta_utility.util import ensure_timezone

if TYPE_CHECKING:
    from collections.abc import Iterable, Sequence
    from typing import Any, Callable, Final

    from eta_utility.type_hints import TimeStep


class RetryWaiter:
    """Helper class which keeps track of waiting time before retrying a connection."""

    VALUES: Final[list[int]] = [0, 1, 3, 5, 5, 10, 20, 30, 40, 60]

    def __init__(self) -> None:
        self.counter = 0

    def tried(self) -> None:
        """Register a retry with the RetryWaiter."""
        self.counter += 1

    def success(self) -> None:
        """Register a successful connection with the RetryWaiter."""
        self.counter = 0

    @property
    def wait_time(self) -> int:
        """Return the time to wait for."""
        if self.counter >= len(self.VALUES) - 1:
            return self.VALUES[-1]
        return self.VALUES[self.counter]

    def wait(self) -> None:
        """Wait/sleep synchronously."""
        sleep(self.wait_time)

    async def wait_async(self) -> None:
        """Wait/sleep asynchronously - must be awaited."""
        await async_sleep(self.wait_time)


def decode_modbus_value(
    value: Sequence[int], byteorder: str, type_: Callable | None = None, wordorder: str = "big"
) -> Any:
    r"""Method to decode incoming modbus values. Strings are always decoded as utf-8 values. If you do not
    want this behaviour specify 'bytes' as the data type.

    :param value: Current value to be decoded into float.
    :param byteorder: Byteorder for decoding i.e. 'little' or 'big' endian.
    :param type\_: Type of the output value. See `Python struct format character documentation
                  <https://docs.python.org/3/library/struct.html#format-characters>` for all possible
                  format strings (default: f).
    :return: Decoded value as a python type.
    """
    if byteorder not in ("little", "big"):
        raise ValueError(f"Specified an invalid byteorder: '{byteorder}'")
    if wordorder not in ("little", "big"):
        raise ValueError(f"Specified an invalid wordorder: '{wordorder}'")

    bo = "<" if byteorder == "little" else ">"

    # Swap words if word order is little endian
    if type_ in (int, float) and wordorder == "little":
        value = value[::-1]

    dtype, _len = _get_decode_params(value, type_)

    # Boolean values don't need decoding
    if type_ is bool:
        return bool(value[0])

    # Determine the format strings for packing and unpacking the received byte sequences. These format strings
    # depend on the endianness (determined by bo), the length of the value in bytes and the data type.
    pack = f">{len(value):1d}H"
    unpack = f"{bo}{_len}{dtype}"

    # Convert the value into the appropriate format
    val = struct.unpack(unpack, struct.pack(pack, *value))[0]
    if type_ is str:
        try:
            val = type_(val, "utf-8")
        except UnicodeDecodeError:
            val = ""
    elif type_ is not None:
        val = type_(val)
    else:
        val = float(val)

    return val


def _get_decode_params(value: Sequence[int], type_: Callable | None = None) -> tuple[str, int]:
    if type_ is str or type_ is bytes:
        dtype = "s"
        _len = len(value) * 2
    elif type_ is bool:
        dtype = "?"
        _len = 1
        if _len != len(value):
            raise ValueError(f"The length of the received value ({len(value)})does not match the data type {type_}")
    elif type_ is int:
        _int_types = {1: "b", 2: "h", 4: "i", 8: "q"}
        _len = 1
        try:
            dtype = _int_types[len(value) * 2]
        except KeyError:
            raise ValueError(
                f"The length of the received value ({len(value)})does not match the data type {type_}"
            ) from None
    elif type_ is float or type_ is None:
        _float_types = {2: "e", 4: "f", 8: "d"}
        _len = 1
        try:
            dtype = _float_types[len(value) * 2]
        except KeyError:
            raise ValueError(
                f"The length of the received value ({len(value)}) does not match the data type: {type_}"
            ) from None
    else:
        raise ValueError(f"The given modbus data type was not recognized: {type_}")

    return dtype, _len


def encode_bits(
    value: str | float | bytes, byteorder: str, bit_length: int, type_: Callable | None = None
) -> list[int]:
    r"""Method to encode python data type to modbus value. This means an array of bytes to send to a
    modbus server.

    :param value: Current value to be decoded into float.
    :param byteorder: Byteorder for decoding i.e. 'little' or 'big' endian.
    :param bit_length: Length of the value in bits.
    :param type\_: Type of the output value. See `Python struct format character documentation
                  <https://docs.python.org/3/library/struct.html#format-characters>` for all possible
                  format strings (default: f).
    :return: Decoded value as a python type.
    """
    byte_length = bit_length // 8
    # Make sure that value is of the type specified by the node.
    if type_ is not None:
        value = type_(value)

    if isinstance(value, int):
        _types = {1: "b", 2: "h", 4: "i", 8: "q"} if value < 0 else {1: "B", 2: "H", 4: "I", 8: "Q"}
        try:
            _type = _types[byte_length]
        except KeyError as e:
            raise ValueError(f"Byte length for integers must be either 1, 2, 4 or 8. Got {byte_length}.") from e
        _len: str | int = ""

    elif isinstance(value, float):
        _types = {2: "e", 4: "f", 8: "d"}
        try:
            _type = _types[byte_length]
        except KeyError as e:
            raise ValueError(f"Byte length for floats must be either 4 or 8. Got {byte_length}.") from e
        _len = ""

    else:
        _type = "s"
        _len = byte_length
        if not isinstance(value, bytes):
            value = bytes(value, "utf-8")

    _order = {"big": ">", "little": "<"}
    try:
        bo = _order[byteorder]
    except KeyError:
        raise ValueError(f"Unknown byte order specified: {byteorder}") from None

    try:
        byte = struct.pack(f"{bo}{_len}{_type}", value)
    except struct.error as e:
        raise ValueError(f"Could not convert value {value!r} to bits.") from e

    bitstrings = [f"{bin(x)[2:]:0>8}" for x in byte]
    return [int(z) for z in "".join(bitstrings)]


def bitarray_to_registers(bits: list[int | bool]) -> list[int]:
    """Convert a list of bits into a list of 16 bit 'bytes'."""
    # Make sure that _bits is a list of integers, not bools.
    _bits = [int(x) for x in bits] if isinstance(bits[0], bool) else bits

    b_size = (len(_bits) + 15) // 16
    register_list = [0] * b_size
    for i in range(0, b_size):
        start = i * 16
        register_list[i] = int("".join([str(v) for v in _bits[start : start + 16]]), 2)

    return register_list


def all_equal(iterable: Iterable[Any]) -> bool:
    """Check if all values inside iterable are equal

    :param iterable: python iterable
    :return: True if all values are equal False elsewhere
    """

    g = groupby(iterable)
    return bool(next(g, True)) and not bool(next(g, False))


class IntervalChecker:
    """Class for the subscription interval checking."""

    def __init__(self) -> None:
        #: Dictionary that stores the value and the time for checking changes and the time interval
        self.node_latest_values: dict[Node, list] = {}

        #: :py:func:`eta_utility.util.ensure_timezone`
        self._assert_tz_awareness = ensure_timezone

    def push(
        self,
        node: Node,
        value: Any | pd.Series | Sequence[Any],
        timestamp: datetime | pd.DatetimeIndex | TimeStep | None = None,
    ) -> None:
        """Push value and time in dictionary for a node. If the value doesn't change compared to the previous
        timestamp, the push is skipped.

        :param node: Node to check.
        :param value: Value from the subscription.
        :param timestamp: Time of the incoming value of the node.
        """

        if node in self.node_latest_values:
            if value != self.node_latest_values[node][0]:
                self.node_latest_values[node] = [value, timestamp]
        elif node.interval is not None:
            self.node_latest_values[node] = [value, timestamp]

    def check_interval_connection(self) -> bool | None:
        """Check the interval between old and new value. If no interval has been defined, the check interval is skipped.

        :return: Boolean for the interval check.
        """

        # Get the current time to compare the interval
        time = self._assert_tz_awareness(datetime.now())

        if len(self.node_latest_values) > 0:
            for node in self.node_latest_values:
                _time_since_last_check = (
                    (time - self.node_latest_values[node][1]).total_seconds()
                    if node in self.node_latest_values
                    else None
                )
                if node in self.node_latest_values and _time_since_last_check is not None:
                    if _time_since_last_check <= float(node.interval):  # type: ignore
                        _changed_within_interval = True
                    else:
                        _changed_within_interval = False
                        break
                else:
                    _changed_within_interval = True
        else:
            _changed_within_interval = True

        return _changed_within_interval



from __future__ import annotations

from typing import TYPE_CHECKING

import torch as th
from stable_baselines3.common import policies

if TYPE_CHECKING:
    from typing import Any


class NoPolicy(policies.BasePolicy):
    """No Policy allows for the creation of agents which do not use neural networks. It does not implement any of
    the typical policy functions but is a simple interface that can be used and ignored. There is no need
    to worry about the implementation details of policies.
    """

    def forward(self, *args: Any, **kwargs: Any) -> None:
        """No Policy allows for the creation of agents which do not use neural networks. It does not implement any of
        the typical policy functions but is a simple interface that can be used and ignored. There is no need
        to worry about the implementation details of policies.
        """
        raise NotImplementedError("'NoPolicy' should be used only, when predictions are calculated otherwise.")

    # type ignored because mypy doesn't seem to think the following is equivalent to the super class...
    def _predict(self, observation: th.Tensor, deterministic: bool = False) -> th.Tensor:  # type: ignore
        """Get the action according to the policy for a given observation.

        Not implemented in NoPolicy.

        :param observation: Observations of the agent.
        :param deterministic: Whether to use stochastic or deterministic actions.
        :return: Taken action according to the policy.
        """
        raise NotImplementedError("'NoPolicy' should be used only, when predictions are calculated otherwise.")

    # type ignored because mypy doesn't seem to think the following is equivalent to the super class...
    def state_dict(  # type: ignore
        self, *, destination: dict[str, Any] | None = None, prefix: str = "", keep_vars: bool = False
    ) -> dict[str, Any]:
        """Returns a dictionary containing a whole state of the module. The dictionary is empty in NoPolicy.

        :param destination: If provided, the state will be updated into the dictionary and the same object returned.
        :param prefix: Prefix added to parameter and buffer names when composing keys in state_dict.
        :param keep_vars: Determine, which variables will be detached from torch.autograd.
        :return: Dictionary with the module/policy state.
        """
        return {}

    # type ignored because mypy doesn't seem to think the following is equivalent to the super class...
    def load_state_dict(self, state_dict: dict[str, Any] | None = None, strict=True) -> None:  # type: ignore
        """Loads the state dictionary. Since the dictionary is always empty, this method doesn't do anything in
        NoPolicy.

        state_dict (dict): a dict containing parameters and
                persistent buffers.
        strict (bool, optional): whether to strictly enforce that the keys
                in :attr:`state_dict` match the keys returned by this module's
                :meth:`~torch.nn.Module.state_dict` function. Default: ``True``
        """
        pass



from __future__ import annotations

from collections.abc import Iterable
from datetime import datetime, timedelta
from typing import TYPE_CHECKING

import pandas as pd

from eta_utility.connectors.base_classes import Connection, SubscriptionHandler
from eta_utility.connectors.modbus import ModbusConnection
from eta_utility.connectors.node import Node, NodeEmonio, NodeModbus
from eta_utility.util import get_logger

if TYPE_CHECKING:
    from collections.abc import Mapping
    from typing import Any

    from eta_utility.type_hints import Nodes, TimeStep

log = get_logger("connectors.emonio")


class EmonioConnection(Connection[NodeEmonio], protocol="emonio"):
    """
    Thin wrapper class for the Emonio that uses a modbus TCP Connection.
    Internally the Emonio nodes are converted to modbus nodes with
    fixed parameters, expect for the name, url and channel.
    If nodes have specified a phase, the connection will check if the phase is connected.
    Additionally, the connection will check for Emonio errors and warnings (max. every minute).

    When creating a :class:`~node.NodeEmonio` the :attr:`~node.NodeEmonio.parameter` (and resulting modbus channel)
    is set by the name of the node (case insensitive).
    See `Available Emonio Nodes` for for possible parameter names.
    Alternatively, the :attr:`~node.NodeEmonio.address` (modbus channel) can be set manually.

    The phase is set by the :attr:`~node.NodeEmonio.phase` attribute of the node.
    Possible values are ``a``, ``b``, ``c`` or ``abc``, with ``abc`` being the default.
    """

    def __init__(self, url: str, *, nodes: Nodes[NodeEmonio] | None = None, check_error: bool = True) -> None:
        """
        Initialize the Emonio connection.

        :param url: URL of the Emonio.
        :param nodes: List of nodes to connect to.
        :param check_error: Setting to check for errors and warnings.
        """
        self.nodes_factory = NodeModbusFactory(url)
        # The actual (Modbus) connection which will be used for reading
        self._connection = ModbusConnection(url=url)
        # Dictionary to keep track of the connected phases
        self._phases_connected: dict[str, bool] = self._get_phases_status()

        self.check_error = check_error
        self._last_error_check = datetime.now() - timedelta(minutes=1)
        self.check_warnings_and_errors()

        # This calls self._validate_nodes(nodes) and checks phases and errors/warnings
        super().__init__(url, nodes=nodes)

    @classmethod
    def _from_node(cls, node: NodeEmonio, **kwargs: Any) -> EmonioConnection:
        """
        Create an Emonio connection from a single node.

        :param node: The node to connect to.
        :return: The Emonio connection.
        """
        return super()._from_node(node)

    def read(self, nodes: Nodes[NodeEmonio] | None = None) -> pd.DataFrame:
        """
        Read the values of the selected nodes.
        If nodes is None, all previously selected nodes will be read.

        :param nodes: List of nodes to read from.
        :return: Dataframe with the read values.
        """

        _nodes: set[NodeEmonio] = self._validate_nodes(nodes)
        return self._connection.read(self._prepare_modbus_nodes(_nodes))

    def write(self, values: Mapping[NodeEmonio, Any]) -> None:
        """
        .. warning::
           Not implemented: Writing to Emonio nodes is not supported.
        """
        raise NotImplementedError("Writing to Emonio nodes is not supported")

    def subscribe(
        self, handler: SubscriptionHandler, nodes: Nodes[NodeEmonio] | None = None, interval: TimeStep = 1
    ) -> None:
        """
        Subscribe to the selected nodes. The Modbus connection does all the handling.

        :param handler: The handler to subscribe.
        :param nodes: List of nodes to subscribe to.
        :param interval: The interval in seconds to read the values.
        """

        _nodes: set[NodeEmonio] = self._validate_nodes(nodes)
        self._connection.subscribe(handler, self._prepare_modbus_nodes(_nodes), interval)

    def close_sub(self) -> None:
        """Close the subscription of the modus connection."""
        self._connection.close_sub()

    def _prepare_modbus_nodes(self, nodes: Nodes[NodeEmonio]) -> list[NodeModbus]:
        """
        Convert the Emonio nodes to modbus nodes with fixed parameters.
        The wordorder is little because the Emonio uses zero based word indexing.
        All values are 32 bit float holding registers.

        :param nodes: List of Emonio nodes.
        :return: List of modbus nodes (will return empty list when no nodes are passed).
        """
        if not isinstance(nodes, Iterable):
            nodes = {nodes}
        return [self.nodes_factory.get_default_node(node.name, node.address) for node in nodes]

    def _validate_nodes(self, nodes: Nodes[NodeEmonio] | None) -> set[NodeEmonio]:
        """
        Validate the nodes and return the set of nodes.
        If nodes is none, return the previously selected nodes.

        :param nodes: List of nodes to validate.
        :return: Set of nodes.
        """
        _nodes: set[NodeEmonio] = super()._validate_nodes(nodes)
        if self.check_error:
            self.check_warnings_and_errors()
        # Check if the phases of the nodes are connected
        self._check_phases(_nodes)
        return _nodes

    def _get_phases_status(self) -> dict[str, bool]:
        """
        Get connection information about the phases of the Emonio.
        The status is read from the discrete input registers of the Emonio.
        """

        nodes = [self.nodes_factory.get_discrete_input_node(phase, i) for i, phase in enumerate(["a", "b", "c"])]
        # Read the status of the phases from the Emonio
        result = self._connection.read(nodes)
        # Convert the result to a dictionary
        result = result.iloc[0].to_dict()
        log.debug("Connected phases: %s", result)
        return result

    def _check_phases(self, nodes: Nodes[NodeEmonio]) -> None:
        """
        Check the connection status of the Emonio phases.

        :param nodes: List of nodes to check.
        """
        if isinstance(nodes, Node):
            nodes = [nodes]

        for node in nodes:
            phase = node.phase
            if phase == "abc":
                continue
            if not self._phases_connected[phase]:
                raise ValueError(f"Phase '{phase}' is not connected.")

    def check_warnings_and_errors(self) -> None:
        """
        Calls the error and warning check if the last check was more than a minute ago.
        """
        now = datetime.now()
        if self._last_error_check is None:
            pass
        elif now - self._last_error_check < timedelta(minutes=1):
            return
        self._last_error_check = now
        self._check_warnings_and_errors()

    def _check_warnings_and_errors(self) -> None:
        """
        Check the Emonio for errors and warnings.
        This is done by reading the error and warning bits from the Emonio.
        If an error bit is set, a ValueError is raised.
        If a warning bit is set, a warning is logged.
        """
        error_node = self.nodes_factory.get_warnings_errors_node("Error", 1000)
        warning_node = self.nodes_factory.get_warnings_errors_node("Warning", 1001)

        result = self._connection.read([error_node, warning_node])

        # Check binary if corresponding error/warning bit is set
        for name, value in result.items():
            # Convert the value to a binary string and reverse it
            for i, bit in enumerate(bin(value.iloc[0])[:1:-1]):
                if bit == "1":
                    # chr(97) = "a", chr(98) = "b", ...
                    msg = (
                        f"{name} bit '{chr(97+i)}' is set. "
                        f"See https://wiki.emonio.de/de/Emonio_P3 for more information."
                    )
                    if name == "Warning":
                        log.warning(msg)
                    else:
                        log.error(msg)
                        raise ValueError(msg)


class NodeModbusFactory:
    """
    The NodeModbusFactory is a factory class that creates NodeModbus objects
    with fixed parameters, expect: name, url and mb_channel.

    Has to be initialized with the url of the Emonio.

    It's a helper class for the EmonioConnection to create its modbus nodes.
    It also can be used to manually create a NodeModbus object, which has to be read with a ModbusConnection.
    (not recommended, use the EmonioConnection instead)
    """

    def __init__(self, url: str):
        common = {"url": url, "protocol": "modbus", "mb_byteorder": "big"}
        self.default_params = {
            **common,
            "mb_wordorder": "little",
            "mb_register": "holding",
            "dtype": "float",
            "mb_bit_length": 32,
        }
        self.discrete_input_params = {
            **common,
            "mb_register": "discrete_input",
            "dtype": "bool",
            "mb_bit_length": 1,
        }
        self.warnings_errors_params = {
            **common,
            "mb_register": "holding",
            "dtype": "int",
            "mb_bit_length": 16,
        }

    def __return_nodes(self, name: str, channel: int, params: dict[str, Any]) -> NodeModbus:
        return NodeModbus(name=name, **params, mb_channel=channel)

    def get_default_node(self, name: str, channel: int) -> NodeModbus:
        """
        Create a modbus node for reading Emonio values.

        :param name: Name of the node.
        :param channel: Modbus channel of the node. (Emonio address)

        """
        return self.__return_nodes(name, channel, self.default_params)

    def get_discrete_input_node(self, name: str, channel: int) -> NodeModbus:
        """
        Create a modbus node for reading the connection status of the Emonio phases.

        :param name: Name of the node.
        :param channel: Modbus channel of the node. (Emonio address)
        """
        return self.__return_nodes(name, channel, self.discrete_input_params)

    def get_warnings_errors_node(self, name: str, channel: int) -> NodeModbus:
        """
        Create a modbus node for reading the error and warning registers of the Emonio.

        :param name: Name of the node.
        :param channel: Modbus channel of the node. (Emonio address)
        """
        return self.__return_nodes(name, channel, self.warnings_errors_params)
